# From https://githooks.com/#how-do-git-hooks-work
from pathlib import Path

from jinja2 import Environment, PackageLoader
from ruamel.yaml import YAML

from yaghm import log

VALID_HOOKS = [
    "applypatch-msg",
    "pre-applypatch",
    "post-applypatch",
    "pre-commit",
    "prepare-commit-msg",
    "commit-msg",
    "post-commit",
    "pre-rebase",
    "post-checkout",
    "post-merge",
    "pre-receive",
    "update",
    "post-receive",
    "post-update",
    "pre-auto-gc",
    "post-rewrite",
    "pre-push",
]


def require_repo_exists(prepo):
    if prepo.exists():
        return

    log.critical(f"Path does not exist: {prepo}")
    exit(1)


def render_wrapper(stage):
    env = Environment(loader=PackageLoader("yaghm"))
    template = env.get_template("wrapper.jinja2")
    return template.render(stage=stage)


def hooks_path(prepo) -> Path:
    return prepo / ".git/hooks"


def backup_path(wrapper_path):
    return Path(f"{wrapper_path}.yaghm.backup")


def commands_path(phooks, hook):
    return Path(phooks) / f"{hook}.yaghm.commands"


def load_config(config_path: Path):
    if not config_path.exists():
        log.critical(f"File does not exist: {config_path}")
        exit(1)

    yaml = YAML()
    config = yaml.load(config_path.open())
    return config



"""
List hooks which are currently enabled in the repo.

# TODO: Support printing for only one hook type

Usage:
    list (-h | --help)
    list REPO
"""
from pathlib import Path

from docopt import docopt

from yaghm import log
from yaghm.lib import (
    hooks_path,
    VALID_HOOKS,
    require_repo_exists,
    render_wrapper,
    commands_path,
)


def main(argv):
    args = docopt(__doc__, argv=argv)

    log.debug(f"Arguments:\n{args}")

    prepo = Path(args["REPO"])
    require_repo_exists(prepo)

    for hook in VALID_HOOKS:
        phooks = hooks_path(prepo)

        pwrapper = phooks / hook
        if not pwrapper.exists():
            log.debug(f"No file at {pwrapper}")
            continue

        wrapper = pwrapper.read_text()
        if wrapper != render_wrapper(hook):
            log.warning(f"Ignoring unexpected script at {pwrapper}")
            continue

        pcmd = commands_path(phooks, hook)
        if not pcmd.exists():
            log.error(f"{pcmd} does not exist - {pwrapper} might be broken.")
            continue

        print(f"{hook}:\n{'='*(len(hook)+1)}\n{pcmd.read_text()}\n")



"""
A minimal git hook manager. For details, see: https://www.github.com/metov/yaghm

Usage:
    yaghm (-h | --help)
    yaghm [--log LEVEL] enable [<args>...]
    yaghm [--log LEVEL] list [<args>...]
    yaghm [--log LEVEL] disable [<args>...]
    yaghm [--log LEVEL] CMDNAME [<args>...]

Options:
    --log LEVEL  Minimum level of logs to print [default: INFO]
"""
import logging

from docopt import docopt

from yaghm import log, enable, disable, list_hooks, custom_command


def main():
    args = docopt(__doc__, options_first=True)

    if level := args["--log"]:
        log.setLevel(logging.getLevelName(level.upper()))

    log.debug(f"Arguments:\n{args}")

    argv = args["<args>"]
    if args["enable"]:
        enable.main(argv)
    elif args["disable"]:
        disable.main(argv)
    elif args["list"]:
        list_hooks.main(argv)
    else:
        custom_command.main(args["CMDNAME"], argv)



import logging
import coloredlogs

# Set up logging
log = logging.getLogger(__name__)
LOGFMT = "%(programname)s %(filename)s:%(lineno)d %(message)s"

# Seems like coloredlogs allows ajusting level up, but not down -- so we start at DEBUG
coloredlogs.install(fmt=LOGFMT, datefmt="%H:%M:%S", level="DEBUG", logger=log)



"""
Run custom command (if defined) for each hook.

Usage:
    CMDNAME [options]

Options:
    --conf PATH  Use given config instead of yaghm.yml in current working directory
    --dryrun  Only echo what would have been done, don't actually do it
"""
import subprocess
from pathlib import Path

from docopt import docopt

from yaghm import log
from yaghm.lib import VALID_HOOKS, load_config


def main(command_name, argv):
    args = docopt(__doc__, argv=argv)

    log.debug(f"Arguments:\n{args}")

    pconf = Path(args["--conf"] or "yaghm.yml")
    config = load_config(pconf)

    for hook in VALID_HOOKS:
        if hook not in config:
            log.debug(f"No {hook} hook defined, skipping.")
            continue

        cmds = config[hook]
        for c in cmds:
            if not isinstance(c, dict):
                log.info(f"No {command_name} defined for: {c}")
                continue

            if command_name not in c:
                log.info(f"No {command_name} defined for: {c['enable']}")
                continue

            cmd = c[command_name]
            if not args["--dryrun"]:
                res = subprocess.run(cmd, shell=True).returncode
            else:
                print(f"Pretending to run: {cmd}")
                res = 0

            if res == 0:
                log.info(f"Command successful: {cmd}")
            else:
                log.error(f"Command failed: {cmd}")



"""
Disable hooks by deleting the wrapper.

Usage:
    disable (-h | --help)
    disable REPO [options]

Options:
    --dryrun  Only echo what would have been done, don't actually do it.
    -y  Non interactive mode, assume safe defaults instead of prompting user
"""
from pathlib import Path

from docopt import docopt

from yaghm import log
from yaghm.lib import (
    require_repo_exists,
    render_wrapper,
    VALID_HOOKS,
    backup_path,
    hooks_path,
)


def main(argv):
    args = docopt(__doc__, argv=argv)

    log.debug(f"Arguments:\n{args}")

    prepo = Path(args["REPO"])
    require_repo_exists(prepo)

    dryrun = bool(args["--dryrun"])
    noask = bool(args["-y"])

    disable_hooks(prepo, dryrun, noask)


def disable_hooks(prepo, dryrun, noask):
    for hook in VALID_HOOKS:
        wrapper = render_wrapper(hook)
        pwrap = hooks_path(prepo) / hook

        if not pwrap.exists():
            continue

        if wrapper != pwrap.read_text():
            log.warning(f"Skipping unexpected file at: {pwrap}")
            continue

        log.info(f"Deleting: {pwrap}")
        pback = backup_path(pwrap)
        if pback.exists():
            log.info(f"Restoring from backup: {pback}")
            if not dryrun:
                backup_path(pwrap).rename(pwrap)
        else:
            if not dryrun:
                pwrap.unlink()



from distutils.core import setup
from pathlib import Path

from setuptools import find_packages

setup(
    name="yaghm",
    version="0.3.0",
    description="Minimal git hook manager for the command line.",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://www.github.com/metov/yaghm",
    author="Azat Akhmetov",
    author_email="azatinfo@yandex.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"": ["templates/*"]},
    install_requires=[
        "coloredlogs",
        "docopt",
        "GitPython",
        "Jinja2",
        "PyYAML",
        "questionary",
    ],
    entry_points={
        "console_scripts": [
            "yaghm=yaghm.main:main",
        ],
    },
)



"""
Read a list of hooks from a config and enable them on the current repo.

You will be prompted if there is a complicated situation, eg. there's an
existing hook managed by another program or a hook is about to be removed,

Usage:
    enable (-h | --help)
    enable REPO [options]

Options:
    --conf PATH  Use given config instead of yaghm.yml in repo root
    --dryrun  Only echo what would have been done, don't actually do it
    -y  Non interactive mode, assume safe defaults instead of prompting user
"""
import subprocess
from pathlib import Path

import questionary
from docopt import docopt

from yaghm import log
from yaghm.lib import (
    hooks_path,
    VALID_HOOKS,
    require_repo_exists,
    render_wrapper,
    backup_path,
    commands_path,
    load_config,
)


def main(argv):
    args = docopt(__doc__, argv=argv)

    log.debug(f"Arguments:\n{args}")

    prepo = Path(args["REPO"])
    require_repo_exists(prepo)

    conf_override = args["--conf"]
    pconf = Path(conf_override) if conf_override else prepo / "yaghm.yml"
    config = load_config(pconf)

    dryrun = bool(args["--dryrun"])
    noask = bool(args["-y"])

    enable_hooks(prepo, config, dryrun, noask)


def enable_hooks(prepo, config, dryrun, noask):
    hooks = collect_hook_commands(config)
    for stage, hooks in hooks.items():
        make_wrapper(prepo, stage, dryrun)
        make_commands(prepo, stage, hooks, dryrun)


def collect_hook_commands(config):
    all_cmds = {}
    for hook in config:
        if hook not in VALID_HOOKS:
            log.warning(f"Skipping invalid hook: {hook}")
            continue

        cmds = all_cmds.setdefault(hook, [])
        for i in config[hook]:
            if isinstance(i, str):
                cmd = i
            elif isinstance(i, dict):
                cmd = i["enable"]
            else:
                log.warning(f"Ignoring unexpected type: {i}")
                continue

            cmds.append(cmd)

        log.debug(f"Collected {len(cmds)} {hook} hooks.")
    return all_cmds


def make_wrapper(prepo, stage, dryrun):
    phooks = hooks_path(prepo)
    pwrap = phooks / stage
    wrapper = render_wrapper(stage)

    if need_to_backup_wrapper(wrapper, pwrap):
        log.warning(f"There's an unexpected file at: {pwrap}")
        backup_wrapper(pwrap, dryrun)

    log.info(f"Writing wrapper at: {pwrap}")
    if not dryrun:
        pwrap.write_text(wrapper)
        subprocess.call(["chmod", "+x", str(pwrap)])


def need_to_backup_wrapper(wrapper, pwrap):
    if not pwrap.exists():
        log.debug(f"No existing {pwrap}")
        return False

    if pwrap.read_text() == wrapper:
        log.debug(f"Existing {pwrap} is already correct")
        return False

    return True


def backup_wrapper(pwrap, dryrun):
    pbackup = backup_path(pwrap)
    log.warning(f"It will be backed up to: {pbackup}")

    if pbackup.exists():
        msg = f"Overwrite old backup in {pbackup}?"
        if not questionary.confirm(msg, default=False).ask():
            log.critical(f"Can't write backup to {pbackup} because it already exists.")
            exit(1)

    if not dryrun:
        pwrap.rename(pbackup)


def make_commands(prepo, hook, newcmds, dryrun):
    pcmds = commands_path(hooks_path(prepo), hook)

    if pcmds.exists():
        oldcmds = set(s for s in pcmds.read_text().splitlines() if s != "")
        deleted = oldcmds - set(newcmds)
        if deleted:
            log.warning(f"{len(deleted)} hooks will be removed:\n" + "\n".join(deleted))
            if not questionary.confirm("Proceed?", default=True).ask():
                log.critical("Aborted on user request.")
                exit(1)

    log.debug(f"Writing command list to: {pcmds}")
    if not dryrun:
        pcmds.write_text("\n".join(newcmds))
